#!/usr/local/bin/perl
# * ====================================================================
# * Copyright (c) 1995,1996, ISIS/JRC.  All rights reserved.
# *
# * This product was developed within the Institute for Safety, Informatics
# * and Systems of the Joint Research Centre of the European Commission.
# *
# * More information about us; and this software product can
# * be found on http://www.jrc.it.
# *
# * With this file you should have received a file called 'License'.
# *
# * Redistribution and use in source and binary forms, with or without
# * modification, are permitted provided that the following conditions
# * are met:
# *
# * 1. Redistributions of source code must retain the above copyright
# *    notice, this list of conditions and the following disclaimer.
# *
# * 2. Redistributions in binary form must reproduce the above copyright
# *    notice, this list of conditions and the following disclaimer in
# *    the documentation and/or other materials provided with the
# *    distribution.
# *
# * 3. All materials mentioning features or use of this
# *    software must display the following acknowledgment:
# *    "This product includes software developed by ISIS/JRC"
# *
# * 4. The name of the Copyright holder must not be used to
# *    endorse or promote products derived from this software without
# *    prior written permission.
# *
# * 5. Redistributions of any form whatsoever must retain the following
# *    acknowledgment:
# *    "This product includes software developed by ISIS/JRC"
# *
# * THIS SOFTWARE IS PROVIDED BY ISIS/JRC ``AS IS'' AND ANY
# * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL ISIS/JRC OR
# * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
# * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
# * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
# * OF THE POSSIBILITY OF SUCH DAMAGE.
# * ====================================================================
#

package ASN1;

use strict;
use vars qw($VERSION @ISA @EXPORT @EXPORT_OK);

use strict;

require Exporter;

@ISA = qw(Exporter AutoLoader);
# Items to export into callers namespace by default. Note: do not export
# names by default without a very good reason. Use EXPORT_OK instead.
# Do not simply export all your public functions/methods/constants.
@EXPORT = qw(
	
);
$VERSION = '0.03';


use BER;
use ASN1::Standard;

=head1 NAME

ASN1 - ASN.1 encoding/decoding

=head1 SYNOPSIS

   use ASN1;

   $ASN1::debug = 1;
   $ASN1::debug_received = 1;
   $ASN1::debug_sent = 1;
   $ASN1::debug_collapse = 1;

   $ASN1::CONNECTION_TIMEOUT = 300;

   &ASN1::sendPDU($handle, $PDU, $descr, 'BER', 'PDU')
      or die $@;
   $content = &ASN1::receivePDU($handle, $descr, 'BER', 'PDU')
      or die $@;

   $oid = &ASN1::OID($oid):
   $str = &ASN1::OIDdescription($oid);

=head1 DESCRIPTION

Module to create a hash/array structure from a BER or XER encoded message
or file.

=head1 AUTHOR

Nick Hibma <Nick.Hibma\@jrc.it, RIT/ISIS/JRC/EC/It

=head2 Remarks

The function uses alarm() for various time outs in the send and receive
functions.

Operations on files are done using unbuffered I/O (sysread() and syswrite()).
Do _not_ use any of the functions read, print, eof, and friends, unless you are
into heavy pain or wizzardry. See the Camel book.

=head2 ASN1 Structure description

The utility asn12perl converts an ASN1 description into a Perl format. This
format contains all the information that is needed to decode and encode PDU's
automatically into a hash/array a structure.

Identifiers for components are used as the hash keys.

In ASN1 the information for one component can be distributed over various other
components. It is not always clear whether a tag is implicit or explicit, if
the tag and terminating type are not specified in the same block. In this case
the bottom element contains a reference to a second description which possibly
contains another reference, etc. When processing a block for the first time,
the description is collapsed and this information is copied into the bottom
element. 

As an example, see the following structure.

   Hello ::= hello [5] World
   World ::= IMPLICIT INTEGER

converted by asn12perl into

   %Hello = (label=>'hello', tag=>5, class=$BER::CONTEXT, ref=>\%World);
   %World = (implicit=>1, tag=>$BER::INTEGER, class=>$BER::UNIVERSAL,
             encoded_as=>$BER::INTEGER);

When processing an element through the block Hello we need to know the fact that
the element is encoded as an INTEGER and has an implicit tag. collapse()
collapses the reference for us to achieve this. collapse(\%Hello) results in:

   %Hello = (label=>'hello', tag=>5, class=>$BER::CONTEXT, ref=>\%World,
             implicit=>1, encoded_as=>$BER::INTEGER, collapsed=>1);

The hash World is not modified.

=head2 Variables & Functions

=over 3

=item $ASN1::debug

Switches on ASN.1 specific debugging.

=item $ASN1::debug_sent, $ASN1::debug_received

Writes every sent/received PDU to a file in the
current directory.

=item $ASN1::debug_collapse

Writes information with regard to collapsing the description tree
to STDERR. This information could be useful in determining why
certain parts of the hash are not included in the resulting output.

=item $ASN1::CONNECTION_TIMEOUT

Determines the time to wait before failing the write to a socket
during sendPDU.

=cut

$ASN1::CONNECTION_TIMEOUT = 300
   if !$ASN1::CONNECTION_TIMEOUT;

$ASN1::sentPDUcount = 0;
$ASN1::receivedPDUcount = 0;


# collapse_ref does the actual work. $descr is the base element and $ref is
# a reference of an element that has to be collapsed into a the base one.
# if the referenced element is a terminating element, 
#
sub collapse_ref ($$) {
   my ($descr, $ref) = @_;

   warn "\$descr is not a hash, but a $descr"
      if $ASN1::debug and ref $descr ne 'HASH';
   warn "\$ref is not a hash, but a $ref"
      if $ASN1::debug and ref $ref ne 'HASH';

   warn "-- Collapse_ref:\n"
      . "   descr = $descr, (".join(',', %{$descr}).")\n"
      . "   ref = $ref, (".join(',', %{$ref}).")\n"
      if $ASN1::debug_collapse;

   push @{ $descr->{labels} }, $ref->{label}
      if exists $ref->{label};

   # If the current element is marked explicit (is equal to 1), then the
   # reference points to the explicit element
   if ( exists $descr->{explicit} ) {
      $descr->{explicit} = $ref;
      return;
   }

   my $descr_is_terminal = (exists $descr->{tag});
   my $ref_is_terminal = (exists $ref->{tag}
                          or exists $ref->{choice}
                          or exists $ref->{external});

   # If the current element ($descr) and the referenced element
   # are both terminals, $ref must be an explicit element
   if ( $descr_is_terminal and $ref_is_terminal and !$descr->{implicit} ) {
      # We found one terminal stacked on top of another one, without an
      # indication of it being an implicit tag. Get out we don't need any info
      # out of the ref struct.

      $descr->{explicit} = $ref;
      return;
   }

   if ( exists $descr->{tag}
        and exists $descr->{implicit}
        and exists $ref->{tag} ) {
      # we've used the implicit; avoid dropping more than one
      # tag because of implicit.

      delete $descr->{implicit};
   } elsif ( !$descr_is_terminal and exists $ref->{tag} ) {
      # copy the info into the top element

      $descr->{tag} = $ref->{tag};
      $descr->{class} = $ref->{class};
   }


   # XXX Could we do this with a simple copy of all the elements
   #     found at that level? Like a foreach (keys ...)?

   foreach my $field ( qw(encoded_as integer bitstr
                          choice seq seqof set setof
                          implicit explicit) ) {
      $descr->{$field} = $ref->{$field}
         if exists $ref->{$field};
   }

   warn "   result = (".join(',', %{$descr}).")\n"
      if $ASN1::debug_collapse;

   &collapse_ref($descr, $ref->{ref})
      if exists $ref->{ref} and !exists $ref->{collapsed};
}


=item collapse($descr)

Collapses the tag, class, seq, set, choice for the reference
and subreferences of an element.

   $descr    the description element to collapse

See the description of ASN1 Structure definition above.

=cut

sub collapse ($) {
   my ($descr) = @_;

   return
      if $descr->{collapsed};

   push @{ $descr->{labels} }, $descr->{label}
      if exists $descr->{label};

   &collapse_ref($descr, $descr->{ref})
      if exists $descr->{ref};

   if ( $ASN1::debug ) {
      my $warning;
      warn "implicit still set"
         if exists $descr->{implicit} and $warning = 1;
      warn "explicit is not a descr ref"
         if exists $descr->{explicit} and ref $descr->{explicit} ne "HASH" and $warning = 1;

      warn '  ('.join(',' %{ $descr }).')'
         if $warning;
   }

   $descr->{labels} = []      # substitute empty array
      if !$descr->{labels};

   $descr->{collapsed} = 1;
}


=item sendData($handle, $buffer[,$bufferlength])

Sends a message through a file handle. Handles timeouts and checks properly
for the length transmitted. Returns true on success, false otherwise. If false
is returned, a printable error string is returned in $@ as well.

=cut

sub sendData ($$;$) {
   my ($handle, $buffer, $bufferlength) = @_;

   $bufferlength = length($buffer)
      if !defined $bufferlength;

   my $old_SIGALRM = $SIG{ALRM};
   local $SIG{ALRM} = sub { warn 'Timeout during send' };
   alarm($ASN1::CONNECTION_TIMEOUT);
   my $lenwritten = syswrite($handle, $buffer, $bufferlength);
   alarm(0);

   # XXX Should we handle short writes more gracefully?

   if ( !defined($lenwritten) ) {
      $@ = "Error during syswrite, $!";
      return undef;
   } elsif ( $lenwritten != $bufferlength ) {
      $@ = 'Short write';
      return undef;
   }

   return 1;         # OK
}
 

=item sendPDU($handle, $PDU, $descr, $encoding [, $name])

Sends a PDU on the connection specified by $handle. The PDU is
a hash to be encoded with $descr.

   $handle     handle for connection
   $PDU        hash of PDU to be sent
   $descr      description to be used when encoding
   $encoding   name of encoding scheme: 'BER', 'XER'
   $name       human readable description for warnings

Returns undef, or, in case of error, an error message.

Set $ASN1::debug_sent if you want to have all the raw PDU's written
to a file before sending.

=cut

sub sendPDU ($$$$;$) {
   my ($handle, $PDU, $descr, $encoding, $name) = @_;

   $name = 'PDU'
      if ! $name;

   # Encode the packet to be sent with the requested encoding
   #
   eval "require 'ASN1/$encoding.pm'";
   if ( $@ ) {
      $@ = "require 'ASN1/$encoding.pm' failed: ".$@.' for '.$name;
      return undef;
   }

   eval '&ASN1::'.$encoding.'::encode($handle,$PDU, $descr,$name); return $@;';
   if ( $@ ) {
      $@ = "encode $encoding failed: ".$@;
      return undef;
   }

   return 1;
}


=item receivePDU($handle, $descr, $encoding [, $name])

Handle the reception of a PDU.

   $handle     handle for the connection
   $descr      description for the hash to be returned
   $encoding   name of encoding scheme: 'BER', 'XER'
   $name       human readable string, for warning messages

Returns the decoded hash for the PDU received (ref $content eq 'HASH'), or, in
case of error, undef (and in that case a error string in $@).

Set $ASN1::debug_received if you want to have all the raw PDU's written to a
file after receiving them. The debug_read flag ($BER::debug_read and
$XER::debug_read) does have no effect if reading XER PDU's from a file
descriptor (handle internally by expat.

=cut

sub receivePDU {
   my ($handle, $descr, $encoding, $name) = @_;

   $name = 'received PDU'
      unless $name;

   my $hash = undef;

   eval "require 'ASN1/$encoding.pm'";
   if ( $@ ) {
      $@ = "require 'ASN1/$encoding.pm' failed: ".$@.' for '.$name;
      return undef;
   }

   my $error = eval '$hash = &ASN1::'.$encoding.'::decode($handle,$descr,$name);
                     return $@;';
   if ( $@ ) {
      $@ = 'decode failed: '.$@;
      return undef;
   }
   if ( $error ) {
      $@ = 'decode failed: '.$error;
      return undef;
   }

   $@ = undef;
   return $hash;
}


=item OID($oid)

   $oid     in dotted numeric format (e.g. 1.2.3.4.5.6)

Returns the cleaned up version of the OID (removing all whitespace
and non dotted decimal chaaracters).

=cut

sub OID ($) {
   my ($oid) = @_;

   $oid =~ s/[^0-9.]//g;   # remove all non digits/points

   return $oid;
}

=item OIDdescription($oid)

   $oid     in dotted numeric format

Returns the human readable version of the OID if a description for it
has been added to %ASN1::OID_descriptions.

New descriptions can be added as follows:

   %ASN1::OID_descriptions = (%ASN1::OID_descriptions,
      '1.2.840.10003' => 'Z39.50',
      '1.2.840.10003.2' => 'APDU'
   )

=cut

sub OIDdescription ($) {
   my ($oid) = @_;

   $oid =~ s/[^0-9.]//g;

   return ($ASN1::OID_descriptions{$oid}? $ASN1::OID_descriptions{$oid} : $oid);
}

1;

__END__
