#!/usr/local/bin/perl
# * ====================================================================
# * Copyright (c) 1995,1996, ISIS/JRC.  All rights reserved.
# *
# * This product was developed within the Institute for Safety, Informatics
# * and Systems of the Joint Research Centre of the European Commission.
# *
# * More information about us; and this software product can
# * be found on http://www.jrc.it.
# *
# * With this file you should have received a file called 'License'.
# *
# * Redistribution and use in source and binary forms, with or without
# * modification, are permitted provided that the following conditions
# * are met:
# *
# * 1. Redistributions of source code must retain the above copyright
# *    notice, this list of conditions and the following disclaimer.
# *
# * 2. Redistributions in binary form must reproduce the above copyright
# *    notice, this list of conditions and the following disclaimer in
# *    the documentation and/or other materials provided with the
# *    distribution.
# *
# * 3. All materials mentioning features or use of this
# *    software must display the following acknowledgment:
# *    "This product includes software developed by ISIS/JRC"
# *
# * 4. The name of the Copyright holder must not be used to
# *    endorse or promote products derived from this software without
# *    prior written permission.
# *
# * 5. Redistributions of any form whatsoever must retain the following
# *    acknowledgment:
# *    "This product includes software developed by ISIS/JRC"
# *
# * THIS SOFTWARE IS PROVIDED BY ISIS/JRC ``AS IS'' AND ANY
# * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL ISIS/JRC OR
# * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
# * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
# * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
# * OF THE POSSIBILITY OF SUCH DAMAGE.
# * ====================================================================
#
# History
#  1999-06-03  NWH   Added support for EXTERNAL
#  1999-06-25  NWH   Fixed bug in receivePDU; If a file was longer than expected
#                    (like 4096 bytes for a 52 byte PDU) reading didn't stop at
#                    52 bytes
#  1999-07-14  NWH   Fix the returning of error messages. On a encode or decode
#                    error the program returns undef for the whole PDU instead
#                    of a partially encoded PDU

package ASN1::BER;

use strict;

use ASN1::Standard;

=head1 NAME

ASN1::BER - Glue module for ASN1 to BER encoding

=head1 SYNOPSIS

   use ASN1::BER;

   $ASN1::BER::debug = 1;
   $ASN1::BER::debug_stacking = 1;

   XXX decode
   xxx encode

=head1 DESCRIPTION

Module to glue together the BER and ASN1 modules.

See the following ASN1 definition encoded in BER:

   Hello = SEQUENCE {
      ciao [4] IMPLICIT INTEGER,
      world [5] IMPLICIT OCTET STRING
   }

The BER module would produce the following structure from that

   $PDU = { tag => 22,
            class => 0,
            value => [
              { tag => 4,
                class => 2,
                value => 1001,
              },
              { tag => 5,
                class => 2,
                value => 'An octet string',
              }
            ]
          };

and that is converted into

   $pdu = [ ciao => 1001, world => 'An octet string' ];

by this module (ASN::BER).

=item $ASN1::BER::debug

Switch on debugging output and checks for this module

=item $ASN1::BER::debug_stacking

Switch on debugging output for the stacking of the items
in the output.

There are probably more debugging flags, but see the source code for their
effect. They are all caled debug_[a-z]+ .

=cut


# decode_integer
#  Convert from a numeric value to a textual identifier if available.
#
sub decode_integer {
   my ($value, $mapping) = @_;

   if ( exists $mapping->{$value} ) {
      return $mapping->{$value};
   } else {
      warn "Value $value not found in mapping";
      return $value;
   }
}

# decode_bitstr
#  Convert a bitstring into a space separated list of names if available.
#
sub decode_bitstr {
   my ($value, $mapping) = @_;

   my $valuer = '';

   foreach my $bitno ( keys %{ $mapping } ) {
      my $bit = $bitno + 7 - 2*($bitno%8);
      if ( vec($value, $bit, 1) ) {
         $valuer .= $mapping->{$bitno} . ' ';
         warn "-- Bit $bitno set ($mapping->{$bitno})"
            if $ASN1::BER::debug;
      }
   }

   return $valuer;
}
      

sub decode_setof {
   my ($setof, $descr) = @_;

   die "SetOf not yet implemented";
}

sub decode_set {
   my ($set, $descr) = @_;

   die "Set not yet implemented";
}

sub decode_seqof {
   my ($seqof, $descr) = @_;

   my $ar = [];

   foreach my $elem ( @{ $seqof } ) {
      my $value = &decode_elem($elem, $descr);
      push @{ $ar }, $value
         if $value;
   }
   warn "-- End of SeqOf (".($#{$ar}+1)." elements)\n"
      if $ASN1::BER::debug;

   return $ar;
}

sub decode_seq {
   my ($seq, $descr) = @_;
   my @descr2 = @{ $descr };
   my $hash = {};

   # The 2 nested loops below have the following task. We have to evaluate
   # every component in sequence against the description hash.  The inner loop
   # checks the current component against the list of remaining components in
   # the description. If it is not found in that list, it is simply skipped.
   # In this case we have to backtrack the description hash.
   # If it is found the subtree is decoded.
   #

   my $i = 0;
   ELEM: foreach my $elem ( @{ $seq } ) {
      my @descr3 = @descr2;

      DESCR: while ( my $descr2 = shift @descr2 ) {
         my $value = &decode_elem($elem, $descr2);
         if ( defined $value ) {
            if ( ref($value) eq 'HASH' ) {
               $hash = { %{ $hash }, %{ $value } };
            } else {
               $hash->{'Unknown'.($i++)} = $value;
            }
            warn "-- $descr2->{label} is there\n"
               if $ASN1::BER::debug;
            next ELEM;
         }
         warn "-- $descr2->{label} not there\n"
            if $ASN1::BER::debug;
      }
      @descr2 = @descr3; # woops, backtrack
      print "Ignored ".Dumper($elem)
         if $ASN1::BER::debug and eval 'require "Data/Dumper.pm";';
   }

   return $hash;
}

sub decode_choice {
   my ($elem, $descr) = @_;

   foreach my $choice ( @{ $descr } ) {
      my $value = &decode_elem($elem, $choice);
      if ( $value ) {
         if ( ref($choice) ) {
            warn "-- choice '$choice->{label}' taken"
               if $ASN1::BER::debug_stacking;
         } else {
            warn "-- plain value chosen"
               if $ASN1::BER::debug_stacking;
         }
      }
      return $value
         if $value;
   }

   return undef;        # no choice applicable
}

# The subroutine below hardcodes quite a few items. This is bad, but the
# easiest option.

sub decode_external ($) {
   my ($elem) = @_;

   my $elem = &decode_elem($elem, \%ASN1::Standard::EXTERNAL);

   if ( exists $elem->{'direct-reference'} ) {
      my $OID = &ASN1::OID($elem->{'direct-reference'});
      if ( exists $elem->{encoding}->{'single-ASN1-type'} ) {
         if ( ref $ASN1::OID2Structure{$OID} eq 'HASH'  # check whether it could a description
              and %{ $ASN1::OID2Structure{$OID} } ) {   # and whether it contains any elements

            # Tie the subdefinition into EXTERNAL->single-ASN1-definition
            # XXX by creating a new hash (with tag=>0, etc in it, we trigger
            # collapse each time. The problem is that we need the context for the external
            # if we want to properly cache the collapsed description

            my $subdescr = {tag=>0, class=>$BER::CONTEXT, ref=>$ASN1::OID2Structure{$OID}};

            $elem->{encoding}->{'single-ASN1-type'} =
                  &decode_elem($elem->{encoding}->{'single-ASN1-type'}, $subdescr);

         } else {
            warn &ASN1::OIDdescription($OID)
               . ' not available in $ASN1::OID2Structure';
         }
      }

      # In case the encoding in octet-aligned or arbitrary is used, the
      # string is not decoded by BER, so we pass it back in one block.
      # The application can, based on the OID do the decoding itself.
   } else {
      # Without an OID we do not know what is in the record.
      # Indirect-reference is not used as on TCP no such thing
      # as presentation layer negotiation exists (see page 69 and 70
      # of the ASN1 specification, section 36 on EXTERNAL)

      warn "No direct-reference in EXTERNAL, not decoded"
         if $ASN1::BER::debug;
   }

   return $elem;
}

sub decode_elem {
   my ($elem, $descr) = @_;
   my $value = undef;

   &ASN1::collapse($descr);

   # the list of elements below is sorted. The dependencies are as follows:
   #   - choice goes before mismatching tags (no tag element)
   #   - mismatched tag goes before everything but choice
   #   - explicit goes before everything else, it's a level of indirection
   #   - value comes last (as a catch-all)

   if ( defined $descr->{choice} ) {
      warn "== Choice $descr->{label}"
         if $ASN1::BER::debug_stacking;
      $value = &decode_choice($elem, $descr->{choice});
      warn "Unable to decode choice $descr->{label}"
         if $ASN1::BER::debug and !$value;
   } elsif ( $elem->{tag} != $descr->{tag} or $elem->{class} != $descr->{class} ) {
      # mismatched tags
      warn "-- Not applicable: $descr->{label} (tag=$descr->{tag}, class=$descr->{class} vs tag=$elem->{tag}, class=$elem->{class})"
         if $ASN1::BER::debug_stacking;
      return undef;
   } elsif ( defined $descr->{explicit} ) {
      warn "== Explicit $descr->{label}"
         if $ASN1::BER::debug_stacking;
      $value = &decode_elem($elem->{value}->[0], $descr->{explicit});
      warn "Unable to decode explicit $descr->{label}"
         if $ASN1::BER::debug and !$value;
   } elsif ( defined $descr->{seq} ) {
      warn "== Seq $descr->{label}"
         if $ASN1::BER::debug_stacking;
      $value = &decode_seq($elem->{value}, $descr->{seq});
      warn "Unable to decode seq $descr->{label}"
         if $ASN1::BER::debug and !$value;
   } elsif ( defined $descr->{seqof} ) {
      warn "== SeqOf $descr->{label}"
         if $ASN1::BER::debug_stacking;
      $value = &decode_seqof($elem->{value}, $descr->{seqof});
      warn "Unable to decode seqof $descr->{label}"
         if $ASN1::BER::debug and !$value;
   } elsif ( defined $descr->{set} ) {
      warn "== Set $descr->{label}"
         if $ASN1::BER::debug_stacking;
      $value = &decode_set($elem->{value}, $descr->{set});
      warn "Unable to decode set $descr->{label}"
         if $ASN1::BER::debug and !$value;
   } elsif ( defined $descr->{setof} ) {
      warn "== SetOf $descr->{label}"
         if $ASN1::BER::debug_stacking;
      $value = &decode_setof($elem->{value}, $descr->{setof});
      warn "Unable to decode setof $descr->{label}"
         if $ASN1::BER::debug_stacking and !$value;
   } else {
      if ( $elem->{class} == $BER::UNIVERSAL ) {
         warn "encoded_as and encoding in current record do not match! ($descr->{encoded_as} vs. $elem->{tag})"
            if $elem->{tag} != $descr->{encoded_as};
         $value = $elem->{value};            # the element is decoded already in the first pass
         if ( $ASN1::BER::debug_stacking ) {
            my $value2 = substr($value, 0, 50);
            $value2 .= ' ...'
               if length($value) > 50;
            $value2 =~ s/\n/\\n/gs;
            $value2 =~ s/\t/ /g;
            $value2 =~ tr/\x00-\x1f\x80-\xff/./;
            warn "== Value $descr->{label} (encoded as ". &BER::tagstr($elem->{tag}, $BER::UNIVERSAL).") = '$value2'";
         }
      } elsif ( defined $descr->{encoded_as} ) {
         if ( $descr->{encoded_as} == $BER::EXTERNAL ) {
            warn "== External $descr->{label}"
               if $ASN1::BER::debug_stacking;
               
            # we have to distingiush between two cases that create different encodings
            if ( $descr->{tag} == $BER::EXTERNAL and $descr->{class} == $BER::UNIVERSAL ) {
               # hello EXTERNAL     'implicit' tagging of EXTERNAL
               $value = &decode_external($elem);
            } else {
               # 'hello [1] EXTERNAL' == 'hello [1] EXPLICIT EXTERNAL'
               $value = &decode_external($elem->{value}->[0]);
            }
         } else {
            $value = &BER::decode_as($elem, $descr->{encoded_as});
            warn "== Value $descr->{label} (encoded as ". &BER::tagstr($descr->{encoded_as}, $BER::UNIVERSAL).")  = '$value'"
               if $ASN1::BER::debug_stacking;
         }
      } else {
         $value = $elem;
         warn "== Value $descr->{label} (unknown format), passed back verbosely"
            if $ASN1::BER::debug_stacking;
      }

      if ( exists $descr->{bitstr} ) {
         $value = &decode_bitstr($value, $descr->{bitstr});
      } elsif ( exists $descr->{integer} ) {
         $value = &decode_integer($value, $descr->{integer});
      }
   }

   if ( defined $descr->{label} ) {
      return {$descr->{label}=>$value}
   } else {
      return $value;
   }
}


sub encode_integer ($$) {
   my ($value, $mapping) = @_;

   # $mapping is { <value> => <name> }

   foreach my $int ( keys %{ $mapping } ) {
      return $value
         if $value eq $mapping->{$int};
   }
}


sub encode_bitstr ($$) {
   my ($value, $mapping) = @_;

   # the mapping is { <bitnumber> => <name> }
   #
   # value can be either a bitmask, bits specified like '01010111'
   # or a list of names of which the names are used in the mapping.

   my $bit;
   my $bitno = 0;          # bit number
   my $valuer = undef;     # return value

   if ( $value =~ m/^[01]+$/ ) {
      foreach my $bitvalue ( split m//, $value ) {
         my $bit = $bitno + 7 - 2*($bitno%8);
         vec($valuer, $bit, 1) = 1
            if $bitvalue eq '1';
      }
   } elsif ( $value =~ m/[^\s0]/ ) {
      foreach $bitno ( keys %{ $mapping } ) {
         #my $bit  = int($bitno / 8)*8 + 7-int($bitno % 8);
         # below without the division
         my $bit = $bitno + 7 - 2*($bitno%8);
         vec($valuer, $bit, 1) = 1
            if $value =~ s/[,\s]*\b($mapping->{$bitno})\b[,\s]*//;
      }
      warn "Unencoded bit names from supplied value: $value"
         if $value !~ m/^[,\s]*$/;
   } else {
      $valuer = $value;
   }

   $value = $valuer
      if $valuer;       # only do it if we actually created a value
   return $value;
}



sub encode_choice ($$$) {
   my ($label, $elem, $descr) = @_;

   my @ar = ();
   foreach my $choice ( @{ $descr } ) {
      return &ASN1::BER::encode_elem($label, $elem, $choice)
         if $label eq $choice->{label};
      push @ar, $choice->{label};
   }
   warn "-- choice $label not found in ".join(', ', @ar)
      if $ASN1::BER::debug;
   $@ = "Choice $label not found";
   return undef;
}

sub encode_external ($$) {
   my ($label, $elem) = @_;

   my $hash = undef;

   if ( $elem->{'direct-reference'} ) {
      my $OID = &ASN1::OID($elem->{'direct-reference'});
      if ( exists $elem->{encoding}->{'single-ASN1-type'} ) {
         if ( ref $ASN1::OID2Structure{$OID} eq 'HASH'   # check whether it could be a description
              and %{ $ASN1::OID2Structure{$OID} } ) {    # and whether it contains any elements

            my $subdescr = {label=>'single-ASN1-type',
                            ref=>$ASN1::OID2Structure{$OID}};
            my $subelem = &encode_elem('single-ASN1-type',
                                       $elem->{encoding}->{'single-ASN1-type'},
                                       $subdescr);
            $hash = {tag=>$BER::EXTERNAL, class=>$BER::UNIVERSAL,
                     name=>'external', value=>[
                       {tag=>$BER::OID, class=>$BER::UNIVERSAL,
                        name=>'direct-reference', value=>$OID},
                       {tag=>0, class=>$BER::CONTEXT, name=>'encoding',
                        value=>[$subelem]}
                    ]};
         } else {
            warn 'OID '.&ASN1::OIDdescription($OID)
               . ' not found in OID2description';
         }
      } else {
         $hash = &encode_elem('', $elem, \%ASN1::Standard::EXTERNAL);
      }
   } else {
      warn "Dunno how to encode $label as EXTERNAL: ".join(', ', %{ $elem });
   }

   return $hash;
}

sub encode_seq ($$$) {
   my ($label, $elem, $descr) = @_;

   if ( ref($elem) ne 'HASH' ) {
      warn "$label is not a ref to a hash";
      return undef;
   }

   my @ar = ();
   my ($hash, $label2);
   my @descr = @{ $descr };            # make a local copy of the array

   warn "-- encode_seq: $label (".join(', ', keys %{ $elem }).')'
      if $ASN1::BER::debug_stacking;
   while ( my $descr2 = shift @descr ) {
      $hash = undef;

      $label2 = $descr2->{label};
      $hash = &ASN1::BER::encode_elem($label2, $elem->{$label2}, $descr2)
         if defined $elem->{$label2};
      if ( $hash ) {
         push @ar, $hash;
      } elsif ( !$descr2->{optional} ) {
         $@ = "$label2 not encoded while not optional";
         return undef;
      }
   }

   return [ @ar ];
}

sub encode_seqof {
   my ($label, $elem, $descr) = @_;

   if ( ref($elem) ne 'ARRAY' ) {
      $@ = "$label not a reference to an array";
      return undef;
   }

   my $hash;
   my @a = ();
   my $i = 0;

   my @elems = @{ $elem };
   while ( $elem = shift @elems ) {
      $hash = &ASN1::BER::encode_elem('', $elem, $descr);
      # we provide no label as this is a Sequence Of

      if ( !$hash ) {
         $@ = "Element $i of $label failed to encode";
         return undef;
      } else {
         push @a, $hash;
      }

      $i++;
   }

   return [ @a ];
}


sub encode_elem {
   my ($label, $elem, $descr) = @_;

   my $value = undef;
   my $hash = undef;

   &ASN1::collapse($descr);

   my $prnt;

   # Below are two different sets of functions.
   # Choice and External produce a hash, while the others
   # produces a value that needs to be encoded in a hash with the
   # right tags.

   if ( exists $descr->{choice} ) { 
      if ( $label eq $descr->{label} ) {
         # duplicate identifier introduced by the following ASN1 code
         # 'B = hello CHOICE { world BOOLEAN, moon BOOLEAN }'
         # in XML: <hello><world>true</world></hello>
         warn "== Stepping over $label:"
            if $ASN1::BER::debug_stacking and $label;
         return undef
            if ref $elem ne 'HASH';
         my $label2 = (keys %{ $elem })[0];
         return &ASN1::BER::encode_elem($label2, $elem->{$label2}, $descr);
      } else {
         warn "== Choice $label from ".join(',', map { $_->{label} } @{ $descr->{choice} })
            if $ASN1::BER::debug_stacking;
         $hash = &ASN1::BER::encode_choice($label, $elem, $descr->{choice});
      }
   } else {
      if ( $label ne $descr->{label} ) {
         warn "-- Labels mismatch ('$label' vs '$descr->{label}')"
            if $ASN1::BER::debug;
      } elsif ( exists $descr->{explicit} ) {
         warn "== Explicit $label"
            if $ASN1::BER::debug_stacking;
         if ( ref($elem) eq 'HASH' ) {
            my $label2 = (keys %{ $elem })[0];
            $value = &ASN1::BER::encode_elem($label2, $elem->{$label2}, $descr->{explicit});
            $value = [$value]
               if $value;
         } else {
            warn "$label not a ref to a hash";
         }
      } elsif ( defined $descr->{seq} ) {
         warn "== Seq $label"
            if $ASN1::BER::debug_stacking;
         $value = &ASN1::BER::encode_seq($label, $elem, $descr->{seq});
      } elsif ( defined $descr->{seqof} ) {
         warn "== SeqOf $label"
            if $ASN1::BER::debug_stacking;
         $value = &ASN1::BER::encode_seqof($label, $elem, $descr->{seqof});
      } elsif ( defined $descr->{set} ) {
         warn "== Set $label"
            if $ASN1::BER::debug_stacking;
         warn "Set not yet implemented";
      } elsif ( defined $descr->{setof} ) {
         warn "== SetOf $label"
            if $ASN1::BER::debug_stacking;
         warn "SetOf not yet implemented";
      } else {
         if ( $descr->{encoded_as} == $BER::EXTERNAL ) {
            $hash = &encode_external($label, $elem);

            if ( $descr->{tag} != $BER::EXTERNAL or $descr->{class} != $BER::UNIVERSAL ) {
               warn "== External $label (explicitly tagged)"
                  if $ASN1::BER::debug_stacking;
               # explicit tagging of EXTERNAL ( 'hello [1] EXTERNAL' )
               # requires the wrapping of the value in the current tag
               # XXX This should extually be done through explicit;
               #     requires changing of collapse.
               $hash = {tag=>$descr->{tag}, class=>$descr->{class},
                        value=>[$hash], name=>'external_explicit'};
            } else {
               warn "== External $label (implicitly tagged)"
                  if $ASN1::BER::debug_stacking;
            }
         } else {
            warn "== Value $label"
               if $ASN1::BER::debug_stacking;

            # Do conversions for named integers and bit strings
            #
            if ( exists $descr->{integer} ) {
               $value = &encode_integer($elem, $descr->{integer});
            } elsif ( exists $descr->{bitstr} ) {
               $value = &encode_bitstr($elem, $descr->{bitstr});
            } else {
               $value = $elem;
               $value = ''
                  if !$value;
            }
         }
      }
   
      if ( defined $value ) {
         $hash->{value} = $value;
         $hash->{name} = $label
            if $BER::debug;               # only add this if the BER module uses it
         $hash->{encoded_as} = $descr->{encoded_as}
            if defined $descr->{encoded_as}
               and !($descr->{tag} == $descr->{encoded_as}
               and $descr->{class} == $BER::UNIVERSAL);
               # be very conservative in adding fields to keep memeory
               # consumption as low as possible.

         if ( defined $descr->{tag} ) {
            $hash->{tag} = $descr->{tag};
            $hash->{class} = (defined $descr->{class}? $descr->{class}: $BER::UNIVERSAL);
         } elsif ( ref($value) ) {
            $hash->{tag}   = $BER::SEQUENCE;
            $hash->{class} = $BER::UNIVERSAL;
         } else {
            $@ = "No tag specified for $label";
            $hash = undef;
         }
      }
   }
   
   if ( $prnt ) {
      use Data::Dumper;
      $Data::Dumper::Indent = 1;
      print STDERR Dumper($hash);
   }

   warn "-- returning empty element for $label"
      if $ASN1::BER::debug_stacking and !$hash;

   return $hash;
}


=item decode($handle, $descr [, $name]);

Reads a BER record from a file opened for reading. Returns a reference to a
hash using the ASN1 identifiers as hash labels, compliant with the XER
specification.

=item encode($hash, $descr [, $name]);

Writes a BER record to a file opened for writing. The input hash is a
tree of name/value pairs, matching the description.

=item debug = 1;

Switch on debugging output

=cut

sub decode ($$;$) {
   my ($handle, $descr, $name) = @_;

   my ($buflen, $buffer) = (0, undef);
   my ($lenread, $len);
   my $pdu;

   if ( ref $handle eq 'SCALAR' ) {
      $pdu = $handle;
   } else {
      warn "-- Reading packet for $name"
         if $ASN1::BER::debug;

      do {
         local $SIG{ALRM} = sub { warn "Timeout during receive" };
         alarm($ASN1::CONNECTION_TIMEOUT);
         $lenread = sysread($handle, $buffer, 1, $buflen);
         alarm(0);

         if ( !defined($lenread) ) {
            #if ( $! == 54 ) {             # XXX ECONNRESET(), errno.ph is broken
            #   $@ = 'Peer closed connection';
            #   return undef;
            #} else {
               $@ = "Error during sysread for $name, $!";
               return undef;
            #}
         } elsif ( $lenread == 0 ) {
            $@ = "Unexpected EOF for $name (while determining length)";
            return undef;
         }

         $buflen += $lenread;
      } while ( ($len = &BER::determineLength(\$buffer)) == $BER::INVALID_LENGTH );

      if ( $len == $BER::INVALID_LENGTH ) {
         $@ = 'Could not determine length for '.$name;
         return undef;
      } elsif ( $len == $BER::INDEFINITE_LENGTH ) {
         # Length of incoming packet is unknown, so let's give it a shot and
         # see how much we get.
         # FIXME Indefinite length packets are not exactly well handled. Using
         # alarms is not exactly the way to go.
         warn "Indefinite length packet encountered, attempting to receive it";
         local $SIG{ALRM} = sub { warn "Timeout during receive of indefinite length transfer" };
         alarm(15);
         while ( $lenread = sysread($handle, $buffer, 100000, $buflen) ) {
            $buflen += $lenread;
            alarm(15);
         }
         alarm(0);
      } else {
         # The length of the buffer is going to be the length of the BER
         # value ($buflen) plus the length of the tag+length fields ($len).
         $len += $buflen;

         while ( $buflen != $len ) {

            local $SIG{ALRM} = sub { warn 'Timeout during BER::decode' };
            alarm($ASN1::CONNECTION_TIMEOUT);
            $lenread = sysread($handle, $buffer, $len-$buflen, $buflen);
            alarm(0);

            if ( ! defined($lenread) ) {
               $@ = "Error during sysread for $name, $!";
               return undef;
            } elsif ( $lenread == 0 ) {
               $@ = "Unexpected EOF for $name";
               return undef;
            }

            $buflen += $lenread;
         }
      }
      $pdu = \$buffer;
   }

   if ( $ASN1::debug_received ) {
      my $filename = $name.'.received.BER.'.($ASN1::receivedPDUcount++);
      if ( open PDU, '>'.$filename ) {
         print PDU $buffer;
         close PDU;
         warn "-- Wrote received packet to '$filename' (".length($$pdu).'b)';
      } else {
         warn "-- Could not open '$filename', $!";
      }
   }

   $@ = undef;

   my ($hash, $pos) = &BER::decode($pdu);
   warn "-- buffer not completely decoded, only $pos out of $buflen"
      if $pos != $buflen;

   return &ASN1::BER::decode_elem($hash, $descr);
}

sub encode ($$$;$) {
   my ($handle, $hash, $descr, $name) = @_;

   my $hash = &ASN1::BER::encode_elem('', $hash, $descr);

   my $pdu = &BER::encode($hash);

   return undef      # nothing to do
      if !$pdu;

   # dump it to file if requested
   #
   if ( $ASN1::debug_sent ) {
      my $filename = "$name.sent.BER.".($ASN1::sentPDUcount++);
      if ( open PDU, '>'.$filename ) {
         print PDU $pdu;
         close PDU;
         warn "-- Wrote sent packet to '$filename' (".length($pdu).'b)';
      } else {
         warn "-- Could not open '$filename', $!";
      }
   }
 
   if ( ref $handle eq 'SCALAR' ) {
      $$handle = $pdu;
   } else {
      if ( !&ASN1::sendData($handle, $pdu, length($pdu)) ) {
         $@ = "Failed to write BER PDU for $name: " . $@;
         return undef;
      }
   }

   return 1;
}

1;
