#!/usr/local/bin/perl
# * ====================================================================
# * Copyright (c) 1995,1996, ISIS/JRC.  All rights reserved.
# *
# * This product was developed within the Institute for Safety, Informatics
# * and Systems of the Joint Research Centre of the European Commission.
# *
# * More information about us; and this software product can
# * be found on http://www.jrc.it.
# *
# * With this file you should have received a file called 'License'.
# *
# * Redistribution and use in source and binary forms, with or without
# * modification, are permitted provided that the following conditions
# * are met:
# *
# * 1. Redistributions of source code must retain the above copyright
# *    notice, this list of conditions and the following disclaimer.
# *
# * 2. Redistributions in binary form must reproduce the above copyright
# *    notice, this list of conditions and the following disclaimer in
# *    the documentation and/or other materials provided with the
# *    distribution.
# *
# * 3. All materials mentioning features or use of this
# *    software must display the following acknowledgment:
# *    "This product includes software developed by ISIS/JRC"
# *
# * 4. The name of the Copyright holder must not be used to
# *    endorse or promote products derived from this software without
# *    prior written permission.
# *
# * 5. Redistributions of any form whatsoever must retain the following
# *    acknowledgment:
# *    "This product includes software developed by ISIS/JRC"
# *
# * THIS SOFTWARE IS PROVIDED BY ISIS/JRC ``AS IS'' AND ANY
# * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL ISIS/JRC OR
# * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
# * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
# * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
# * OF THE POSSIBILITY OF SUCH DAMAGE.
# * ====================================================================
#

# This file contains two modules which are closely linked. The first one
# is the interface to ASN1.pm
#
# The second one is the implementation of the XML parser on top of
# Expat / XML::Parser.

# History:
# 1999-07-28   NWH   Char discarded new lines, "<tag>hello</tag>" is now
#                    different from "<tag>\nhello\n</tag>", unfortunately.


package ASN1::XER;

=head1 NAME

ASN1::XER - Glue module for ASN1 to XER encoding

=head1 SYNOPSIS

   use ASN1::XER;

   $ASN1::XER::debug = 1;
   $ASN1::XER::STREAMDELIMITER = '^L';

   decode($handle, $descr, 'decodable packet')
      or die $@;
   encode($handle, $hash, $descr, 'decodable packet');
      or die $@;

=head1 DESCRIPTION

Module to do the decoding of incoming XER encoded message.

=head2 Variables & Functions

=over 3

=item $ASN1::XER::debug, $ASN1::XER::debug_stacking, $ASN1::XER::debug_encoding

Switch on debugging output and checks for this module:

   debug_stacking    the stacking of elements
   debug_encoding    the encoding of the elements
   debug             switches on checks and warnings for possible mistakes

=item $ASN1::XER::STREAMDELIMITER

If multiple documents are sent across a connection, this string
sits in between aline on a line.

=cut


$ASN1::XER::STREAMDELIMITER = '';

sub encode_value {
   my ($value, $descr) = @_;

   my $label = "[$descr->{tag}/$descr->{class}], encoded as [$descr->{encoded_as}/0]";

   if ( $ASN1::XER::debug ) {
      warn "tag is missing"
         if !exists $descr->{tag};
      warn "encoded_as is missing"
         if !exists $descr->{encoded_as};
   }
   warn "-- $label = '".substr($value, 0, 40)."'"
      if $ASN1::XER::debug_encoding;

   if ( !exists $descr->{encoded_as} ) {
      warn "non-value expected but found value";
      if ( $ASN1::XER::debug ) {
         warn '  descr: '.join(',', %{ $descr });
         warn '  value: \''.$value.'\'';
      }
      return undef;
   }

   my $encoding = $descr->{encoded_as};

   if ( $encoding == $BER::NULL ) {
      return '';
   } elsif ( $encoding == $BER::INTEGER ) {
      if ( exists $descr->{integer}          # if it is an enumerated value
           and $value =~ m/^\d+$/            # value is a number (index into enumeration hash)
           and exists $descr->{integer}->{$value} ) {    # and the number exists in the hash
         return $descr->{integer}->{$value};
      } else {
         return $value;                      # if not, pass verbosely XXX correct?
      }
   } elsif ( $encoding == $BER::OID ) {
      return &ASN1::OID($value);
   } elsif (  $encoding == $BER::REAL ) {
      return $value;
   } elsif ( $encoding == $BER::BOOLEAN ) {
      # XER requires either 'false' or 'true'. Convert common equivalents
      return ($value =~ m/^(0|off|false)$/i? 'false':'true');
   } elsif ( $encoding == $BER::BIT_STR ) {
      if ( exists $descr->{bitstr}
           and $value =~ m/^[01]+$/ ) {
         my @bits = split m//, $value;
         my $bitno = 0;
         my $bit;
         my @strings = ();
         while ( $bit = pop @bits ) {
            push @strings, $descr->{bitstr}->{$bitno}
               if $bit eq '1';
            $bitno++;
         }
         return join(' ', @strings);
      } else {
         $value =~ tr/,\x08 / /s;   # convert space, comma and tab into one space
         return $value;             # reverse the order
      }
   } elsif ( $encoding == $BER::OCTET_STR
             or $encoding == $BER::UTF8STR
             or $encoding == $BER::NUM_STR
             or $encoding == $BER::PRN_STR
             or $encoding == $BER::TELETEX
             or $encoding == $BER::VIDEOTEX
             or $encoding == $BER::IA5_STR
             or $encoding == $BER::UTCTIME
             or $encoding == $BER::GENTIME
             or $encoding == $BER::GRP_STR
             or $encoding == $BER::VIS_STR
             or $encoding == $BER::GEN_STR
             or $encoding == $BER::UNI_STR
             or $encoding == $BER::BMP_STR
           ) {
      if ( $value =~ m/[^\x08\x0a\x0d\x20-\x7f]/ ) {
         # woops, high ascii/control characters; we'll have to cast it in hex
         my $rvalue = ' ';
         my $i;
         for ($i = 0; $i < length($value); $i++) {
            # FIXME SLOOOW! Has to be fixed someday. Let's get it working first.
            $rvalue .= sprintf("%02x ", ord(substr($value, $i, 1)));
         }
         return '<Hex>'.$rvalue.'</Hex>';
      } else {
         $value =~ s/\&/\&amp;/g;
         $value =~ s/\</\&lt;/g;
         $value =~ s/\>/\&gt;/g;
         $value =~ s/\"/\&quot;/g;
         $value =~ s/\'/\&apos;/g;

         return $value;
      }
   } else {
      warn "Unknown encoding $encoding for ".$label;
   }

   return $value;
}

sub encode_choice ($$$) {
   my ($label, $elem, $descr) = @_;

   my $subdescr;

   if ( ref $descr ne 'ARRAY' ) {
      warn "Expected array for description, but found ".ref $descr;
      return undef;
   }

   foreach $subdescr ( @{ $descr } ) {
      next
         if !exists $subdescr->{label};

      return &encode_elem($label, $elem, $subdescr)
         if $label eq $subdescr->{label};
   }
   warn "choice '$label' not found";

   return undef;
}

sub encode_explicit ($$$) {
   my ($label, $elem, $descr) = @_;

   my $sublabel = (keys %{ $elem })[0];

   return &encode_elem($sublabel, $elem, $descr);
}

sub encode_seq ($$$) {
   my ($label, $elem, $descr) = @_;

   if ( ref $elem ne 'HASH' ) {
      warn "Expected ref to hash for $label, but found ".$elem;
      return;
   }
   if ( ref $descr ne 'ARRAY' ) {
      warn "Expected array for description for $label, but found ".$descr;
      return;
   }

   my @buffer = ();

   # keep track of the labels we've done to see missing ones.
   my ($sublabel, @sublabels) = ();

   # Iterate through the list of elements of the sequence
   my $subdescr;
   foreach $subdescr ( @{ $descr } ) {
      warn "Element of sequence in $label without label"
         if $ASN1::XER::debug and !exists $subdescr->{label};

      $sublabel = $subdescr->{label};
      if ( exists $elem->{$sublabel} ) {
         push @buffer, &encode_elem($sublabel, $elem, $subdescr);
         push @sublabels, $sublabel;
      } else {
         warn "$sublabel missing from $label, but not optional"
            if !exists $subdescr->{optional};
      }
   }

   # check the once which were not in the description of the sequence
   # but were in the presented hash, the combination of the error
   # message above and this one should give the spelling mistakes
   # rightaway.
   my $sublabels = join('|', @sublabels);
   my @rest = grep !/$sublabels/, keys %{ $elem };
   warn "Ignored labels: ".join(',', @rest)."\n  from ".join(',', map { $_->{label} } @{ $descr })
      if @rest;

   return join("\n", grep {$_} @buffer);
}

sub encode_seqof ($$$) {
   my ($label, $elem, $descr) = @_;

   if ( ref $elem ne 'ARRAY' ) {
      warn "Expected array for $label, but found ".$elem;
      if ( eval 'require "Data/Dumper.pm"' ) {
         $Data::Dumper::Indent = 1;
         print STDERR &Dumper($elem);
      }
      return;
   }

   my @buffer = ();
   my $i = 0;
   foreach my $subelem ( @{ $elem } ) {
      my $sublabel = $label.($i++);
      push @buffer, '<Item>'.&encode_elem($sublabel, $subelem, $descr).'</Item>';
   }

   return join("\n", grep {$_} @buffer);
}

sub encode_external ($) {
   my ($elem) = @_;

   my $buffer = undef;

   if ( $elem->{'direct-reference'} ) {
      my $OID = &ASN1::OID($elem->{'direct-reference'});
      if ( exists $elem->{encoding}->{'single-ASN1-type'} ) {
         if ( ref $ASN1::OID2Structure{$OID} eq 'HASH'   # check whether it could be a description
              and %{ $ASN1::OID2Structure{$OID} } ) {    # and whether it contains any elements

            my $subdescr = {label=>'single-ASN1-type', ref=>$ASN1::OID2Structure{$OID}};
            $buffer = &encode_elem('single-ASN1-type', $elem->{encoding}, $subdescr)
         } else {
            warn 'OID '.&ASN1::OIDdescription($OID)
               . ' not available as a record syntax';
            warn "Either the package was not 'use'd or OID is not in $ASN1::OID2Structure"
               if $ASN1::XER::debug;
         }
         $buffer = '<direct-reference>'.$OID.'</direct-reference>'
                 . "<encoding>\n".$buffer."\n</encoding>";
      } else {
         $buffer = &encode_elem('', $elem, \%ASN1::Standard::EXTERNAL);
      }
   } else {
      warn "Dunno how to encode $label as EXTERNAL: ".join(', ', %{ $elem });
   }

   return "<External>\n".$buffer."\n</External>";
}

sub encode_elem ($$$) {
   my ($label, $elem, $descr) = @_;

   if ( ref $elem eq 'HASH') {
      warn "== label: '$label' value: '$elem->{$label}'\n   descr: $descr, (".join(',', %{ $descr }).')'
         if $ASN1::XER::debug_encoding;
   } else {
      warn "== label: '$label' value: '$elem'\n   descr: $descr, (".join(',', %{ $descr }).')'
         if $ASN1::XER::debug_encoding;
   }

   &ASN1::collapse($descr);

   my @labels = @{ $descr->{labels} };

   my @rest;      # keep track of leftovers for debugging

   # Try to match the labels from the description ($descr and sub
   # references 'ref=>[]', with the labels specified in the hash
   # ASN1 allows for mutliple labels for one terminal:
   # One ::= SEQUENCE {
   #         two  [1] Two;
   # }
   # Two ::= CHOICE {
   #         three INTEGER;
   #         four  BOOLEAN;
   # }
   # should be
   # $hash->{one}->{two}->{three} = 1;

   my ($before, $after) = ('', '');
   my $descrlabel;

   while ( $descrlabel = shift @labels ) {
      if ( $label ne $descrlabel ) {
         warn "Wrong label, expected '$descrlabel', got '$label'";
         return undef;
      }
      $before .= "<$label>";
      $after = "</$label>$after";

      $elem = $elem->{$label};
      if ( ref $elem eq 'HASH' ) {
         ($label, @rest) = keys %{ $elem };
      } else {
         warn 'Ignored required labels \''.join(',', @labels).'\''
            if $ASN1::XER::debug and @labels;
         $label = undef;
      }

      last
         if !@labels;

      warn "Ignored extra labels '".join(',', @rest)."', apart from $label"
         if $ASN1::XER::debug and @rest;
   }

   return undef
      if !defined $elem and $descr->{optional};

   # Get the subelem if there is one. There should be none or one.
   my ($subelemkey, @rest) = grep m/seq|set|seqof|setof|choice|explicit/, keys %{ $descr };
   warn "Found extraneous ".join(',', @rest)." in $descr->{label}"
      if $ASN1::XER::debug and @rest;
   warn "-- Subelement for $label is $subelemkey"
      if $ASN1::XER::debug_stacking and $subelemkey;

   if ( !$subelemkey ) {
      if ( $descr->{encoded_as} == $BER::EXTERNAL ) {
         my $buffer = &encode_external($elem);
         return $before."\n".$buffer."\n".$after;
      } else {
         my $buffer = &encode_value($elem, $descr);
         return $before.$buffer.$after;
      }
   } else {
      my $buffer;
      if ( $subelemkey eq 'choice' ) {
         $buffer = &encode_choice($label, $elem, $descr->{choice});
      } elsif ( $subelemkey eq 'explicit' ) {
         $buffer = &encode_explicit($label, $elem, $descr->{explicit});
      } elsif ( $subelemkey eq 'seq' ) {
         $buffer = &encode_seq($label, $elem, $descr->{seq});
      } elsif ( $subelemkey eq 'seqof' ) {
         $buffer = &encode_seqof($label, $elem, $descr->{seqof});
      } elsif ( $subelemkey eq 'set' ) {
         #$buffer = &encode_set($label, $elem, $descr->{set});
         warn 'XXX SET not implemented yet';
      } elsif ( $subelemkey eq 'setof' ) {
         #$buffer = &encode_setof($label, $elem, $descr->{setof});
         warn 'XXX SET OF not implemented yet';
      } else {
         warn "Unknown subelemkey '$subelemkey'";
      }
      if ($before) {
         return $before."\n".$buffer."\n".$after;
      } else {
         return $buffer;
      }
   }
}

=item decode($handle, $descr [, $name]);

Reads a XER record from a file opened for reading. Returns a reference to a
hash using the ASN1 identifiers as hash labels, compliant with the XER
specification.

XXX the decoding process completely ignores the description. No checking not
conversion is done yet to correct the resulting hash yet.

=cut

sub decode ($$;$) {
   my ($handle, $descr, $name) = @_;

   my $parser = new XML::Parser(Style=>'XER', ErrorContext=>3);

   my $hash = undef;
   if ( ref $handle eq 'SCALAR' ) {
      eval { $hash = $parser->parse($$handle."\n") };

      if ( $ASN1::debug_received ) {
         my $filename = $name.'.received.XER.'.($ASN1::receivedPDUcount++);
         if ( open PDU, '>'.$filename ) {
            print PDU $$handle;
            close PDU;
            warn "-- Wrote received packet to '$filename' (".length($$handle).'b)';
         } else {
            warn "-- Could not open '$filename', $!";
         }
      }
   } else {
      eval { $hash = $parser->parse($handle) };

      # XXX if we let XML::Parser read the file we cannot dump its
      # contents to a debugging file
   }

   $@ = undef     # not an error, simply no data
      if $@ =~ m/no element found at line 1, column 0, byte -1/;

   if ( exists $hash->{Search} ) {
      $hash = $hash->{Search};
   } else {
      $@ = 'Tag \'Search\' not found among '.join(', ', keys %{ $hash });
   }

   if ($@) {
      $@ = "Invalid XML received for $name\n" . $@;
      $hash = undef;
   }

   return $hash;
}

=item encode($handle, $hash, $descr [, $name]);

Writes a XER record to a file opened for writing. The input hash must be formed
according to the XER specification (a reference to a hash received from a
previous decode will do just fine).

=cut

sub encode ($$$;$) {
   my ($handle, $hash, $descr, $name) = @_;

   if ( keys %{ $hash } > 1 ) {
      warn "Too many keys in hash";
      return;
   }

   if ( !$descr ) {
      $@ = 'no description given';
      return undef;
   }

   my $label = (keys %{ $hash })[0];
   my $pdu = &ASN1::XER::encode_elem($label, $hash, $descr);

   $pdu = "<Search>\n".$pdu."\n</Search>\n";

   # dump it to file if requested
   #
   if ( $ASN1::debug_sent ) {
      my $filename = "$name.sent.XER.".($ASN1::sentPDUcount++);
      if ( open PDU, '>'.$filename ) {
         print PDU $pdu;
         close PDU;
         warn "-- Wrote sent packet to '$filename' (".length($pdu).'b)';
      } else {
         warn "-- Could not open '$filename', $!";
      }
   }

   if ( ref $handle eq 'SCALAR' ) {
      $$handle = $pdu;
   } else {
      if ( !&ASN1::sendData($handle, $pdu, length($pdu)) ) {
         $@ = "Failed to write XER PDU for $name: " . $@;
         return undef;
      }

      my $delimiter = "$ASN1::XER::STREAMDELIMITER\n";
      if ( !&ASN1::sendData($handle, $delimiter, length($delimiter)) ) {
         $@ = "Failed to write XER delimiter: " . $@;
         return undef;
      }
   }

   return 1;
}

1;


package XML::Parser::XER;

use strict;
use XML::Parser;
$XML::Parser::Built_In_Styles{XER} = 1;

use ASN1::Standard;

# Special features of the XER encoding
# - tags are unique within a scope -> hash possible
# - char elements consist of one piece only
# - a tag contains either a subtree or a char element
# - Item tags indicate an array of elements
# - External, Item, and any other tag with an initial capital
#   should be removed as they are introduced by the encoding.
#
# All these features are exploited to make sure we get a tree
# that is as small as possible.

sub Init {
   my ($expat) = @_;

   $expat->{StartDocument} = undef;          # start of document tag
   $expat->{Hashes} = [];                    # stack of hierarchies
   $expat->{CurElem} = $expat->{XER} = {};   # root element
   $expat->{Stream_Delimiter} = $ASN1::XER::STREAMDELIMITER;
}

sub Start {
   my ($expat, $tag) = @_;

   $expat->{StartDocument} = $tag
      if !$expat->{StartDocument};

   # Push the surrounding element onto the stack
   push @{ $expat->{Hashes} }, $expat->{CurElem};
   $expat->{CurElem} = undef;
}

sub Char {
   my ($expat, $text) = @_;

   # XXX string concatenation is slow for big chunks.
   # should be replaced by push, but this is difficult to
   # decode in End below.
   $expat->{CurElem} .= $text       # ignore any text if there is also a subelement
      unless ref $expat->{CurElem};
}

sub End {
   my ($expat, $tag) = @_;

   my $content = $expat->{CurElem};
   $content = ''                 # delete the content if it only consists of whitespace
      unless $content =~ m/\S/;

   # Pop surrounding element off the stack
   $expat->{CurElem} = pop @{ $expat->{Hashes} };

   # zap any char element if it was added (conserve memory)
   $expat->{CurElem} = undef
      unless ref $expat->{CurElem};

   if ( $tag eq 'Item' ) {
      # Item indicates a list of items
      push @{ $expat->{CurElem} }, $content;
   } elsif ( $tag eq 'External' ) {
      # Ignore tag
      $expat->{CurElem} = $content;
   } elsif ( $tag eq 'Hex' ) {
      # Hexadecimal encoding of the contents
      $content = pack('c*', map { hex($_) } grep {$_} split m/\s+/, $content);
      $expat->{CurElem} = $content;
   } else {
      $expat->{CurElem}->{$tag} = $content;
   }
}

sub Final {
   my ($expat) = @_;

   # clean up the kitchen
   delete $expat->{CurElem};
   delete $expat->{Hashes};

   # and get the children to school
   $expat->{XER};
}

1;
